@@ -1,3 +1,4 @@
+pub mod manager;
 mod maps;
 pub mod protocol;
 
@@ -5,22 +6,22 @@ use std::{
     iter,
     pin::Pin,
     sync::{
-        atomic::{self, AtomicBool},
         Arc,
+        atomic::{self, AtomicBool},
     },
     task::Poll,
     time::Duration,
 };
 
 use futures_core::{Future, Stream};
-use futures_util::{future::join_all, SinkExt, StreamExt};
+use futures_util::{SinkExt, StreamExt, future::join_all};
 use parking_lot::Mutex;
 use thiserror::Error;
 use tokio::{
     select,
     sync::{
-        mpsc::{self, UnboundedReceiver},
         Semaphore,
+        mpsc::{self, UnboundedReceiver},
     },
     task::JoinHandle,
 };
@@ -28,18 +29,26 @@ use tokio_tungstenite::tungstenite;
 use tungstenite::error::UrlError;
 use url::Url;
 
-use self::maps::*;
-use self::protocol::*;
+use self::{
+    maps::*,
+    protocol::{Message, MessageOrRequest, Request, WebsocketMessage, WebsocketRequest},
+};
 
 use crate::{
-    socket,
-    util::{keep_flushing, CancelOnDrop, TimeoutOnDrop},
-    Error,
+    Error, socket,
+    util::{CancelOnDrop, TimeoutOnDrop, keep_flushing},
 };
 
 type WsMessage = tungstenite::Message;
 type WsError = tungstenite::Error;
-type WsResult<T> = Result<T, tungstenite::Error>;
+type WsResult<T> = Result<T, Error>;
+type GetUrlResult = Result<Url, Error>;
+
+impl From<WsError> for Error {
+    fn from(err: WsError) -> Self {
+        Error::failed_precondition(err)
+    }
+}
 
 const WEBSOCKET_CLOSE_TIMEOUT: Duration = Duration::from_secs(3);
 
@@ -48,11 +57,11 @@ const PING_TIMEOUT: Duration = Duration::from_secs(3);
 
 const RECONNECT_INTERVAL: Duration = Duration::from_secs(10);
 
-pub struct Response {
+struct Response {
     pub success: bool,
 }
 
-pub struct Responder {
+struct Responder {
     key: String,
     tx: mpsc::UnboundedSender<WsMessage>,
     sent: bool,
@@ -78,8 +87,8 @@ impl Responder {
         })
         .to_string();
 
-        if let Err(e) = self.tx.send(WsMessage::Text(response)) {
-            warn!("Wasn't able to reply to dealer request: {}", e);
+        if let Err(e) = self.tx.send(WsMessage::Text(response.into())) {
+            warn!("Wasn't able to reply to dealer request: {e}");
         }
     }
 
@@ -101,7 +110,7 @@ impl Drop for Responder {
     }
 }
 
-pub trait IntoResponse {
+trait IntoResponse {
     fn respond(self, responder: Responder);
 }
 
@@ -132,7 +141,7 @@ where
     }
 }
 
-pub trait RequestHandler: Send + 'static {
+trait RequestHandler: Send + 'static {
     fn handle_request(&self, request: Request, responder: Responder);
 }
 
@@ -156,8 +165,10 @@ impl Stream for Subscription {
 fn split_uri(s: &str) -> Option<impl Iterator<Item = &'_ str>> {
     let (scheme, sep, rest) = if let Some(rest) = s.strip_prefix("hm://") {
         ("hm", '/', rest)
-    } else if let Some(rest) = s.strip_suffix("spotify:") {
+    } else if let Some(rest) = s.strip_prefix("spotify:") {
         ("spotify", ':', rest)
+    } else if s.contains('/') {
+        ("", '/', s)
     } else {
         return None;
     };
@@ -169,7 +180,7 @@ fn split_uri(s: &str) -> Option<impl Iterator<Item = &'_ str>> {
 }
 
 #[derive(Debug, Clone, Error)]
-pub enum AddHandlerError {
+enum AddHandlerError {
     #[error("There is already a handler for the given uri")]
     AlreadyHandled,
     #[error("The specified uri {0} is invalid")]
@@ -186,7 +197,7 @@ impl From<AddHandlerError> for Error {
 }
 
 #[derive(Debug, Clone, Error)]
-pub enum SubscriptionError {
+enum SubscriptionError {
     #[error("The specified uri is invalid")]
     InvalidUri(String),
 }
@@ -224,8 +235,23 @@ fn subscribe(
     Ok(Subscription(rx))
 }
 
+fn handles(
+    req_map: &HandlerMap<Box<dyn RequestHandler>>,
+    msg_map: &SubscriberMap<MessageHandler>,
+    uri: &str,
+) -> bool {
+    if req_map.contains(uri) {
+        return true;
+    }
+
+    match split_uri(uri) {
+        None => false,
+        Some(mut split) => msg_map.contains(&mut split),
+    }
+}
+
 #[derive(Default)]
-pub struct Builder {
+struct Builder {
     message_handlers: SubscriberMap<MessageHandler>,
     request_handlers: HandlerMap<Box<dyn RequestHandler>>,
 }
@@ -267,22 +293,26 @@ impl Builder {
         subscribe(&mut self.message_handlers, uris)
     }
 
+    pub fn handles(&self, uri: &str) -> bool {
+        handles(&self.request_handlers, &self.message_handlers, uri)
+    }
+
     pub fn launch_in_background<Fut, F>(self, get_url: F, proxy: Option<Url>) -> Dealer
     where
-        Fut: Future<Output = Url> + Send + 'static,
-        F: (FnMut() -> Fut) + Send + 'static,
+        Fut: Future<Output = GetUrlResult> + Send + 'static,
+        F: (Fn() -> Fut) + Send + 'static,
     {
         create_dealer!(self, shared -> run(shared, None, get_url, proxy))
     }
 
-    pub async fn launch<Fut, F>(self, mut get_url: F, proxy: Option<Url>) -> WsResult<Dealer>
+    pub async fn launch<Fut, F>(self, get_url: F, proxy: Option<Url>) -> WsResult<Dealer>
     where
-        Fut: Future<Output = Url> + Send + 'static,
-        F: (FnMut() -> Fut) + Send + 'static,
+        Fut: Future<Output = GetUrlResult> + Send + 'static,
+        F: (Fn() -> Fut) + Send + 'static,
     {
         let dealer = create_dealer!(self, shared -> {
             // Try to connect.
-            let url = get_url().await;
+            let url = get_url().await?;
             let tasks = connect(&url, proxy.as_ref(), &shared).await?;
 
             // If a connection is established, continue in a background task.
@@ -303,15 +333,47 @@ struct DealerShared {
 }
 
 impl DealerShared {
-    fn dispatch_message(&self, msg: Message) {
+    fn dispatch_message(&self, mut msg: WebsocketMessage) {
+        let msg = match msg.handle_payload() {
+            Ok(value) => Message {
+                headers: msg.headers,
+                payload: value,
+                uri: msg.uri,
+            },
+            Err(why) => {
+                warn!("failure during data parsing for {}: {why}", msg.uri);
+                return;
+            }
+        };
+
         if let Some(split) = split_uri(&msg.uri) {
-            self.message_handlers
+            if self
+                .message_handlers
                 .lock()
-                .retain(split, &mut |tx| tx.send(msg.clone()).is_ok());
+                .retain(split, &mut |tx| tx.send(msg.clone()).is_ok())
+            {
+                return;
+            }
         }
+
+        debug!("No subscriber for msg.uri: {}", msg.uri);
     }
 
-    fn dispatch_request(&self, request: Request, send_tx: &mpsc::UnboundedSender<WsMessage>) {
+    fn dispatch_request(
+        &self,
+        request: WebsocketRequest,
+        send_tx: &mpsc::UnboundedSender<WsMessage>,
+    ) {
+        trace!("dealer request {}", &request.message_ident);
+
+        let payload_request = match request.handle_payload() {
+            Ok(payload) => payload,
+            Err(why) => {
+                warn!("request payload handling failed because of {why}");
+                return;
+            }
+        };
+
         // ResponseSender will automatically send "success: false" if it is dropped without an answer.
         let responder = Responder::new(request.key.clone(), send_tx.clone());
 
@@ -325,13 +387,11 @@ impl DealerShared {
             return;
         };
 
-        {
-            let handler_map = self.request_handlers.lock();
+        let handler_map = self.request_handlers.lock();
 
-            if let Some(handler) = handler_map.get(split) {
-                handler.handle_request(request, responder);
-                return;
-            }
+        if let Some(handler) = handler_map.get(split) {
+            handler.handle_request(payload_request, responder);
+            return;
         }
 
         warn!("No handler for message_ident: {}", &request.message_ident);
@@ -355,9 +415,9 @@ impl DealerShared {
     }
 }
 
-pub struct Dealer {
+struct Dealer {
     shared: Arc<DealerShared>,
-    handle: TimeoutOnDrop<()>,
+    handle: TimeoutOnDrop<Result<(), Error>>,
 }
 
 impl Dealer {
@@ -376,6 +436,14 @@ impl Dealer {
         subscribe(&mut self.shared.message_handlers.lock(), uris)
     }
 
+    pub fn handles(&self, uri: &str) -> bool {
+        handles(
+            &self.shared.request_handlers.lock(),
+            &self.shared.message_handlers.lock(),
+            uri,
+        )
+    }
+
     pub async fn close(mut self) {
         debug!("closing dealer");
 
@@ -383,7 +451,7 @@ impl Dealer {
 
         if let Some(handle) = self.handle.take() {
             if let Err(e) = CancelOnDrop(handle).await {
-                error!("error aborting dealer operations: {}", e);
+                error!("error aborting dealer operations: {e}");
             }
         }
     }
@@ -402,7 +470,7 @@ async fn connect(
     let default_port = match address.scheme() {
         "ws" => 80,
         "wss" => 443,
-        _ => return Err(WsError::Url(UrlError::UnsupportedUrlScheme)),
+        _ => return Err(WsError::Url(UrlError::UnsupportedUrlScheme).into()),
     };
 
     let port = address.port().unwrap_or(default_port);
@@ -455,13 +523,13 @@ async fn connect(
                 Ok(close_frame) => ws_tx.send(WsMessage::Close(close_frame)).await,
                 Err(WsError::AlreadyClosed) | Err(WsError::ConnectionClosed) => ws_tx.flush().await,
                 Err(e) => {
-                    warn!("Dealer finished with an error: {}", e);
+                    warn!("Dealer finished with an error: {e}");
                     ws_tx.send(WsMessage::Close(None)).await
                 }
             };
 
             if let Err(e) = result {
-                warn!("Error while closing websocket: {}", e);
+                warn!("Error while closing websocket: {e}");
             }
 
             debug!("Dropping send task");
@@ -484,19 +552,19 @@ async fn connect(
                     Some(Ok(msg)) => match msg {
                         WsMessage::Text(t) => match serde_json::from_str(&t) {
                             Ok(m) => shared.dispatch(m, &send_tx),
-                            Err(e) => info!("Received invalid message: {}", e),
+                            Err(e) => warn!("Message couldn't be parsed: {e}. Message was {t}"),
                         },
                         WsMessage::Binary(_) => {
                             info!("Received invalid binary message");
                         }
                         WsMessage::Pong(_) => {
-                            debug!("Received pong");
+                            trace!("Received pong");
                             pong_received.store(true, atomic::Ordering::Relaxed);
                         }
                         _ => (), // tungstenite handles Close and Ping automatically
                     },
                     Some(Err(e)) => {
-                        warn!("Websocket connection failed: {}", e);
+                        warn!("Websocket connection failed: {e}");
                         break;
                     }
                     None => {
@@ -517,12 +585,15 @@ async fn connect(
                 timer.tick().await;
 
                 pong_received.store(false, atomic::Ordering::Relaxed);
-                if send_tx.send(WsMessage::Ping(vec![])).is_err() {
+                if send_tx
+                    .send(WsMessage::Ping(bytes::Bytes::default()))
+                    .is_err()
+                {
                     // The sender is closed.
                     break;
                 }
 
-                debug!("Sent ping");
+                trace!("Sent ping");
 
                 sleep(PING_TIMEOUT).await;
 
@@ -556,8 +627,9 @@ async fn run<F, Fut>(
     initial_tasks: Option<(JoinHandle<()>, JoinHandle<()>)>,
     mut get_url: F,
     proxy: Option<Url>,
-) where
-    Fut: Future<Output = Url> + Send + 'static,
+) -> Result<(), Error>
+where
+    Fut: Future<Output = GetUrlResult> + Send + 'static,
     F: (FnMut() -> Fut) + Send + 'static,
 {
     let init_task = |t| Some(TimeoutOnDrop::new(t, WEBSOCKET_CLOSE_TIMEOUT));
@@ -575,13 +647,13 @@ async fn run<F, Fut>(
                     () = shared.closed() => break,
                     r = t0 => {
                         if let Err(e) = r {
-                            error!("timeout on task 0: {}", e);
+                            error!("timeout on task 0: {e}");
                         }
                         tasks.0.take();
                     },
                     r = t1 => {
                         if let Err(e) = r {
-                            error!("timeout on task 1: {}", e);
+                            error!("timeout on task 1: {e}");
                         }
                         tasks.1.take();
                     }
@@ -593,12 +665,12 @@ async fn run<F, Fut>(
                         break
                     },
                     e = get_url() => e
-                };
+                }?;
 
                 match connect(&url, proxy.as_ref(), &shared).await {
                     Ok((s, r)) => tasks = (init_task(s), init_task(r)),
                     Err(e) => {
-                        error!("Error while connecting: {}", e);
+                        error!("Error while connecting: {e}");
                         tokio::time::sleep(RECONNECT_INTERVAL).await;
                     }
                 }
@@ -609,4 +681,6 @@ async fn run<F, Fut>(
     let tasks = tasks.0.into_iter().chain(tasks.1);
 
     let _ = join_all(tasks).await;
+
+    Ok(())
 }
